-- Normalizing If-Expressions
-- Author:: Andreas Abel
-- Created :: 2005-07-06 
-- Source  :: IfExpressions.agda

-- Cut out data type NIfExpr and Bits
-- Use just framework

-- # import Property

Bool : Set
true  : Bool
false : Bool
if    : Bool -> Bool -> Bool -> Bool

-- Axiom:: rotation for if 

rot_if : (c,ct,ce,t,e:Bool) -> 
    if (if c ct ce) t e == if c (if ct t e) (if ce t e)

-- proof elsewhere

-- Fix some set of variables and some valuation

A : Set
f : A -> Bool

-- If expressions over variables in some set A

Expr : Set
Var : A -> Expr
If  : Expr -> Expr -> Expr -> Expr

-- Semantics

eval : Expr -> Bool
evalVar : (a:A) -> eval (Var a) == f a
evalIf  : (c,t,e:Expr) -> eval (If c t e) == if (eval c) (eval t) (eval e)

axEval : ((a:A) -> eval (Var a) == f a)
       /\ ((c,t,e:Expr) -> eval (If c t e) == if (eval c) (eval t) (eval e))
-- Normalization of if expressions

-- composition of normal expressions

nif : Expr -> Expr -> Expr -> Expr
nifVar : (x:A) -> (t,e:Expr) -> nif (Var x) t e == If (Var x) t e
nifIf  : (a,b,c,t,e:Expr) -> nif (If a b c)  t e == If a (nif b t e) (nif c t e)
  -- invariant:: a is Var x
axNif : ((x:A) -> (t,e:Expr) -> nif (Var x) t e == If (Var x) t e)
      /\ ((a,b,c,t,e:Expr) -> nif (If a b c)  t e == If a (nif b t e) (nif c t e))

-- normalization

norm : Expr -> Expr
normVar : (x:A) -> norm (Var x) == Var x
normIf  : (c,t,e:Expr) -> norm (If c t e) == nif (norm c) (norm t) (norm e)

axNorm : ((x:A) -> norm (Var x) == Var x)
       /\ ((c,t,e:Expr) -> norm (If c t e) == nif (norm c) (norm t) (norm e))

-- Soundness of normalization ----------------------------------------

-- Induction for Expr

indExpr : (P : Expr -> Prop) ->
            ((a:A) -> P (Var a)) ->
            ((c,t,e : Expr) -> P c -> P t -> P e -> P (If c t e)) ->
            (e : Expr) -> P e

-- Lemma:: soundness of nif

{- Lemma:: eval (nif c t e) == eval (If c t e)
          (emb implicit)
          
Proof:: By induction on c.

Case NIf:: 
  LHS     eval (nif (If a ct ce) t e)
       == eval (If a (nif ct t e) (nif ce t e))
       == if (f a) (eval (nif ct t e)) (eval (nif ce t e))
  (IH) == if (f a) (eval (If ct t e)) (eval (If (ce t e))
       == if (f a) (if (eval ct) (eval t) (eval e))
                   (if (eval ce) (eval t) (eval e))
  RHS     eval (If (If a ct ce) t e)
       == if (eval (If a ct ce)) (eval t) (eval e)
       == if (if (f a) (eval ct) (eval ce)) (eval t) (eval e)
 rot_if== LHS
-}

Sound_nif : Expr -> Prop
Sound_nif = \ c -> (t,e : Expr) ->
                 (eval (If c t e))
              == (eval (nif c t e))

sound_nif  : (c : Expr) -> Sound_nif c
sound_nif = indExpr Sound_nif
                  fol-plugin(axEval,axNif)    -- just reflexivity
                  (\ c t e ihc iht ihe -> fol-plugin(ihc,iht,ihe,rot_if,axEval,axNif))


-- Theorem:: soundness of norm

Sound_norm : (e:Expr) -> Prop
Sound_norm = \ e -> (eval e) 
                 == (eval (norm e))

sound_norm : (e:Expr) -> Sound_norm e
sound_norm = indExpr Sound_norm
               (fol-plugin(axEval,axNorm))     -- just reflexivity
               (\ c t e ihc iht ihe -> fol-plugin (ihc,iht,ihe,sound_nif,axEval,axNorm))
