
-- # import prelude
-- # import prop
-- # import id
-- # import nat
-- # import fin
-- # import eqchain

{--------------------------------------------------------------------------
    The terms
 --------------------------------------------------------------------------}

infixl 20 @

data Term : Nat -> Set where
  var (n : Nat) : Fin n -> Term n
  (@) (n : Nat) : Term n -> Term n -> Term n
  λ (n : Nat) : Term (suc n) -> Term n

{--------------------------------------------------------------------------
    Raising a term
 --------------------------------------------------------------------------}

shiftFrom (n:Nat) : Fin (suc n) -> Term n -> Term (suc n)
shiftFrom i (var j)	= var (thin i j)
shiftFrom i (t @ u)	= shiftFrom i t @ shiftFrom i u
shiftFrom i (λ t)	= λ (shiftFrom (fsuc i) t)

shift (n:Nat) : Term n -> Term (suc n)
shift = shiftFrom fzero

shiftCommute (n : Nat) :
    (i : Fin (suc n)) -> (j : Fin (suc (suc n))) -> (t : Term n) ->
    shiftFrom j (shiftFrom i t) ==
    shiftFrom (thin j i) (shiftFrom (thick i j) t)
shiftCommute i j (var k) = cong var (thinCommute k i j)
shiftCommute i j (s @ t) = cong2 (@) (shiftCommute i j s) (shiftCommute i j t)
shiftCommute i j (λ t)   = cong λ (shiftCommute (fsuc i) (fsuc j) t)

shiftCommute0 (n : Nat) :
    (i : Fin (suc n)) -> (t : Term n) ->
    shiftFrom fzero (shiftFrom i t) ==
    shiftFrom (fsuc i) (shiftFrom fzero t)
shiftCommute0 i t with finNonZero i =>
	      i t nonZero = shiftCommute i fzero t


{--------------------------------------------------------------------------
    Substitution
 --------------------------------------------------------------------------}

substitute (n:Nat) : Term n -> Fin (suc n) -> Term (suc n) -> Term n
substitute u i (var j) with thinView i j =>
    u i (var i) hit		  = u
    u i (var (thin i k)) (miss k) = var k
substitute u i (s @ t)		  = substitute u i s @ substitute u i t
substitute u i (λ t)		  = λ (substitute (shift u) (fsuc i) t)

-- Substituting after shifting --------------------------------------------

substShift (n:Nat) : (i : Fin (suc n)) -> (t : Term n) -> (s : Term n) -> 
		     substitute s i (shiftFrom i t) == t
substShift i (var j) s = rewrite thinViewMiss i j in refl
substShift i (t @ u) s = rewrite substShift i t s
				 substShift i u s
			 in refl
substShift i (λ t) s   = rewrite substShift (fsuc i) t (shiftFrom fzero s)
			 in refl

-- shiftSubst (n:Nat) :
--     (j : Fin (suc n)) ->
--     (s : Term n) -> (i : Fin (suc n)) -> (u : Term (suc n)) ->
--     shiftFrom j (substitute s i u) ==
--     substitute (shiftFrom j s) (thin (thin (inj i) j) i) (shiftFrom (thin (inj i) j) u)
-- shiftSubst j s i (var k) with thinView i k =>
--     j s i (var  i	 )  hit      =
-- 	rewrite thinViewHit (thin (thin (inj i) j) i) in refl
--     j s i (var (thin i k)) (miss k)  = subst Goal (sym $ thinCommute k i B) lem
--       where
-- 	B = thin (inj i) j
-- 
-- 	Goal : Fin _ -> Set
-- 	Goal z = var (thin j k) == substitute (shiftFrom j s)
-- 					      (thin B i) (var z)
-- 	lem : Goal (thin (thin B i) (thin (thick i B) k))
-- 	lem = rewrite thickThin i j
-- 		      thinViewMiss (thin B i) (thin j k)
-- 	      in refl
-- shiftSubst j s i (t @ u) = cong2 (@) (shiftSubst j s i t) (shiftSubst j s i u)
-- shiftSubst j s i (λ u) with finNonZero j =>
--   j s i (λ u) nonZero = cong λ $ subst Goal (sym $ shiftCommute j fzero s) lem
--     where
--       Goal : Term _ -> Set
--       Goal z = shiftFrom (fsuc j) (substitute (shiftFrom fzero s) (fsuc i) u) ==
-- 		substitute z (fsuc (thin (thin (inj i) j) i))
-- 			     (shiftFrom (fsuc (thin (inj i) j)) u)
-- 
--       lem : Goal (shiftFrom (fsuc j) (shiftFrom fzero s))
--       lem = shiftSubst (fsuc j) (shiftFrom fzero s) (fsuc i) u

shiftSubst (n : Nat) :
    (D : Fin (suc (suc n))) ->
    (i : Fin (suc n)) -> (s : Term n) -> (u : Term (suc n)) ->
    shiftFrom (thick i D) (substitute s i u)
    ==
    substitute (shiftFrom (thick i D) s) (thin D i) (shiftFrom D u)
shiftSubst D i s (var k) with thinView i k =>
  D i s (var i) hit = rewrite thinViewHit (thin D i) in refl
  D i s (var (thin i k)) (miss k) = subst Goal (sym $ thinCommute k i D) lem
    where
      Goal : Fin _ -> Set
      Goal z = var (thin (thick i D) k) ==
	       substitute (shiftFrom (thick i D) s) (thin D i) (var z)

      lem : Goal (thin (thin D i) (thin (thick i D) k))
      lem = rewrite thinViewMiss (thin D i) (thin (thick i D) k) in refl
shiftSubst D i s (t @ u) = cong2 (@) (shiftSubst D i s t)
				     (shiftSubst D i s u)
shiftSubst D i s (λ u) =
    cong λ
    $ subst (\z -> shiftFrom (fsuc (thick i D)) (substitute (shift s) (fsuc i) u)
		== substitute z (fsuc (thin D i)) (shiftFrom (fsuc D) u))
	    (sym $ shiftCommute0 (thick i D) s)
    $ shiftSubst (fsuc D) (fsuc i) (shift s) u

shiftSubst0 (n : Nat) :
    (s : Term n) -> (i : Fin (suc n)) -> (u : Term (suc n)) ->
    shiftFrom fzero (substitute s i u) ==
    substitute (shiftFrom fzero s) (fsuc i) (shiftFrom fzero u)
shiftSubst0 s i u with finNonZero i =>
    s i u nonZero = shiftSubst fzero i s u

-- Commutativity of substitution ------------------------------------------

substCommute (n : Nat) :
	     (s : Term n) -> (u : Term (suc n)) -> (t : Term (suc (suc n))) ->
	     (i : Fin (suc n)) -> (j : Fin (suc (suc n))) ->
    substitute (substitute s i u) (thick i j)
	       (substitute (shiftFrom (thick i j) s) (thin j i) t)
    == substitute s i (substitute u j t)
substCommute s u (var k) i j with thinView j k =>
   s u (var j) i j hit		  =
      rewrite thinViewThick j i
	      thinViewHit (thick i j)
      in refl
   s u (var (thin j k)) i j (miss k) with thinView i k =>
      s u (var (thin j i)) i j (miss i) hit	=
	  rewrite thinViewMiss j i
		  thinViewHit (thin j i)
		  substShift (thick i j) s
	  in refl
      s u (var (thin j (thin i k))) i j (miss (thin i k)) (miss k) =
	  rewrite thinViewMiss j (thin i k)
	  in subst (\x -> LHS x == var k) (sym $ thinCommute k i j) lem
	    where
	      LHS : Fin _ -> Term _
	      LHS h = substitute (substitute s i u) (thick i j)
				 (substitute (shiftFrom (thick i j) s)
					     (thin j i)
					     (var h)
				 )

	      lem : LHS (thin (thin j i) (thin (thick i j) k)) == var k
	      lem =
		rewrite thinViewMiss (thin j i) (thin (thick i j) k)
			thinViewMiss (thick i j) k
		in refl

substCommute s u (t1 @ t2) i j	= cong2 (@) (substCommute s u t1 i j) 
					    (substCommute s u t2 i j)
substCommute s u (λ t) i j	=
    cong λ
    $ subst (\z -> Goal z rhs2) lem1
    $ subst (\z -> Goal lhs1 z) lem2
    $ substCommute (shift s) (shift u) t (fsuc i) (fsuc j)
  where
    lhs : (n : Nat) -> Term n -> Fin _ -> Term _ -> Term _
    lhs _ s0 i0 u0 = substitute (shiftFrom fzero s0) (fsuc i0) (shiftFrom fzero u0)
    rhs : (n : Nat) -> Term n -> Fin _ -> Term _ -> Term _
    rhs _ s0 i0 u0 = shiftFrom fzero (substitute s0 i0 u0)

    lhs1 = substitute (shiftFrom fzero s) (fsuc i) (shiftFrom fzero u)
    rhs1 = shiftFrom fzero (substitute s i u)

    lhs2 = shiftFrom (fsuc (thick i j)) (shiftFrom fzero s)
    rhs2 = shiftFrom fzero (shiftFrom (thick i j) s)

    lem1 : lhs1 == rhs1
    lem1 = sym $ shiftSubst0 s i u

    lem2 : lhs2 == rhs2
    lem2 = sym $ shiftCommute0 (thick i j) s
    
    Goal : Term _ -> Term _ -> Set
    Goal z1 z2 = substitute z1 (fsuc (thick i j))
		  (substitute z2 (fsuc (thin j i)) t)
	      == substitute (shiftFrom fzero s) (fsuc i)
		  (substitute (shiftFrom fzero u) (fsuc j) t)

substCommute0 (n : Nat) :
	     (s : Term n) -> (u : Term (suc n)) -> (t : Term (suc (suc n))) ->
	     (i : Fin (suc n)) ->
  substitute (substitute s i u) fzero (substitute (shift s) (fsuc i) t)
  == substitute s i (substitute u fzero t)
substCommute0 {n} s u t i =
  rewrite thickZero i in substCommute {n} _ _ _ _ fzero
				    -- ^ annoying

{--------------------------------------------------------------------------
    The reduction relation
 --------------------------------------------------------------------------}

data (-->) {n:Nat} : Term n -> Term n -> Set where
  β     (t : Term (suc n))(u : Term n) : λ t @ u --> substitute u fzero t
  rapp1 (t,t',u : Term n)	       : t --> t' -> t @ u --> t' @ u
  rapp2 (t,u,u' : Term n)	       : u --> u' -> t @ u --> t @ u'
  rλ    (t,t' : Term (suc n))	       : t --> t' -> λ t --> λ t'


{--------------------------------------------------------------------------
    (Reflexive) Transitive closure
 --------------------------------------------------------------------------}

data TC {A:Set}(R:A -> A -> Set) : A -> A -> Set where
  tcrefl (a:A)	   : TC R a a
  tccons (a,b,c:A) : R a b -> TC R b c -> TC R a c

(-->*) (n:Nat) : Term n -> Term n -> Set
t -->* u = TC (-->) t u

tcunit (n:Nat)(a,b:Term n) : a --> b -> a -->* b
tcunit r = tccons r tcrefl

tcapp1 (n:Nat)(t,t',u : Term n) : t -->* t' -> t @ u -->* t' @ u
tcapp1 tcrefl	     = tcrefl
tcapp1 (tccons r tr) = tccons (rapp1 r) (tcapp1 tr)

tcapp2 (n:Nat)(t,u,u' : Term n) : u -->* u' -> t @ u -->* t @ u'
tcapp2 tcrefl	      = tcrefl
tcapp2 (tccons r tr) = tccons (rapp2 r) (tcapp2 tr)

tcλ (n:Nat)(t,t' : Term (suc n)) : t -->* t' -> λ t -->* λ t'
tcλ tcrefl	    = tcrefl
tcλ (tccons r tr)  = tccons (rλ r) (tcλ tr)

tctrans (n:Nat)(s,t,u:Term n) : s -->* t -> t -->* u -> s -->* u
tctrans tcrefl	       r = r
tctrans (tccons r1 r2) r = tccons r1 (tctrans r2 r)

infixr 10 `tccons`, `tccons'`

tccons' (n:Nat)(s,t,u:Term n) : s --> t -> t -->* u -> s -->* u
tccons' r tr = tccons r tr

{--------------------------------------------------------------------------
    Reducing under substitutions
 --------------------------------------------------------------------------}

rsubst1 (n:Nat)(s, s' : Term (suc n))(t : Term n)(i : Fin (suc n)) :
	s --> s'  -> substitute t i s --> substitute t i s'
rsubst1 (rλ r)	  = rλ (rsubst1 r)
rsubst1 (rapp1 r) = rapp1 (rsubst1 r)
rsubst1 (rapp2 r) = rapp2 (rsubst1 r)
rsubst1 {n} {λ t @ u} {substitute u fzero t} {s} {i} β =
    subst (\v -> substitute s i (λ t @ u) --> v)
	  (substCommute0 _ _ _ _) β

rshift (n:Nat)(t, t' : Term n)(i : Fin (suc n)) :
       t --> t' -> shiftFrom i t --> shiftFrom i t'
rshift (rλ r) = rλ (rshift r)
rshift (rapp1 r) = rapp1 (rshift r)
rshift (rapp2 r) = rapp2 (rshift r)
rshift {n} {λ t @ u} {substitute u fzero t} {i} β =
    subst Goal lem β
  where
    Goal : Term (suc n) -> Set
    Goal z = λ (shiftFrom (fsuc i) t) @ shiftFrom i u --> z

    lhs = substitute (shiftFrom i u) fzero (shiftFrom (fsuc i) t)
    rhs = shiftFrom i (substitute u fzero t)

    lem : lhs == rhs
    lem = sym $ shiftSubst (fsuc i) fzero u t

tcsubst2 (n:Nat)(t, t' : Term n)(s : Term (suc n))(i : Fin (suc n)) :
	 t --> t' -> substitute t i s -->* substitute t' i s
tcsubst2 {n} {t} {t'} {var j} {i} r with thinView i j =>
  {n} {t} {t'} {var i}	  {i} r hit	 = tcunit r
  {n} {t} {t'} {var (thin i j)} {i} r (miss j) = tcrefl
tcsubst2 {n} {t} {t'} {s @ u} {i} r =
    tctrans (tcapp1 (tcsubst2 r))
	    (tcapp2 (tcsubst2 r))
tcsubst2 {n} {t} {t'} {λ s} {i} r = tcλ $ tcsubst2 $ rshift r

{--------------------------------------------------------------------------
    Now what do we want to prove?
 --------------------------------------------------------------------------}

LocalConfluence (A:Set) : (A -> A -> Set) -> Set
LocalConfluence {A} R =
  (u, u1, u2 : A)      ->
  R u u1 -> R u u2 ->
  { v : A; TC R u1 v /\ TC R u2 v }

Confluence (A:Set) : (A -> A -> Set) -> Set
Confluence {A} R =
  (u, u1, u2 : A)        ->
  TC R u u1 -> TC R u u2 ->
  { v : A; TC R u1 v /\ TC R u2 v }

-- note: there are some problems with using things like Exists with a higher
-- order parameter, together with the first order unification used in the pattern
-- matching.
localConfluence (n:Nat) : LocalConfluence {Term n} (-->)

-- rapp1 rapp1
localConfluence (t @ u) (t1 @ u) (t2 @ u) (rapp1 r1) (rapp1 r2)
			      with localConfluence t t1 t2 r1 r2 =>
  (t @ u) (t1 @ u) (t2 @ u) (rapp1 r1) (rapp1 r2) (v, andI r1' r2') =
      (v @ u, andI (tcapp1 r1') (tcapp1 r2'))

-- rapp2 rapp2
localConfluence (t @ u) (t @ u1) (t @ u2) (rapp2 r1) (rapp2 r2)
			      with localConfluence u u1 u2 r1 r2 =>
  (t @ u) (t @ u1) (t @ u2) (rapp2 r1) (rapp2 r2) (v, andI r1' r2') =
      (t @ v, andI (tcapp2 r1') (tcapp2 r2'))

-- rapp1 rapp2
localConfluence (t @ u) (t' @ u) (t @ u') (rapp1 r1) (rapp2 r2) =
      (t' @ u', andI (tcapp2 $ tcunit r2) (tcapp1 $ tcunit r1))

-- rapp2 rapp1
localConfluence (t @ u) (t @ u') (t' @ u) (rapp2 r1) (rapp1 r2) =
      (t' @ u', andI (tcapp1 $ tcunit r2) (tcapp2 $ tcunit r1))

-- rλ rλ
localConfluence (λ u) (λ u1) (λ u2) (rλ r1) (rλ r2)
			      with localConfluence u u1 u2 r1 r2 =>
  (λ u) (λ u1) (λ u2) (rλ r1) (rλ r2) (v, andI r1' r2') =
      (λ v, andI (tcλ r1') (tcλ r2'))

-- β β
localConfluence (λ u @ t) (substitute t fzero u) (substitute t fzero u) β β =
  (substitute t fzero u, andI tcrefl tcrefl)

-- β rapp1
localConfluence (λ u @ t) (substitute t fzero u) (λ u' @ t) β (rapp1 (rλ r)) =
  (substitute t fzero u', andI (tcunit $ rsubst1 r) (tcunit β))
-- rapp1 β
localConfluence (λ u @ t) (λ u' @ t) (substitute t fzero u) (rapp1 (rλ r)) β =
  (substitute t fzero u', andI (tcunit β) (tcunit $ rsubst1 r))

-- β rapp2
localConfluence (λ u @ t) (substitute t fzero u) (λ u @ t') β (rapp2 r) =
  (substitute t' fzero u, andI (tcsubst2 r) (tcunit β))
-- rapp2 β
localConfluence (λ u @ t) (λ u @ t') (substitute t fzero u) (rapp2 r) β =
  (substitute t' fzero u, andI (tcunit β) (tcsubst2 r))

confluence (n:Nat) : Confluence {Term n} (-->)

confluence u u  u   tcrefl	  tcrefl	= (u, andI tcrefl tcrefl)
confluence u u' u  (tccons r tr)  tcrefl	= (u', andI tcrefl (tccons' r tr))
confluence u u  u'  tcrefl	 (tccons r tr)	= (u', andI (tccons' r tr) tcrefl)

confluence u u1 u2 (tccons r1 tr1) (tccons r2 tr2)
		   with localConfluence u _ _ r1 r2 =>
  u u1 u2 (tccons r1 tr1) (tccons r2 tr2) (v, andI tr1' tr2')
		   with confluence _ v u2 tr2' tr2 =>
    u u1 u2 (tccons r1 tr1) (tccons r2 tr2) (v, andI tr1' tr2') (v', andI tr2a tr2b)
		   with confluence _ u1 v' tr1 (tctrans tr1' tr2a) =>
      u u1 u2 (tccons r1 tr1) (tccons r2 tr2) (v, andI tr1' tr2')
					      (v', andI tr2a tr2b)
					      (v'', andI tr1a tr2c)
	= (v'', andI tr1a (tctrans tr2b tr2c))

Y (n:Nat) : Term n
Y {n} = λ $ (λ $ f @ (x @ x)) @ (λ $ f @ (x @ x))
  where
    f : Term (2 + n)
    f = var (fsuc fzero)

    x : Term (2 + n)
    x = var fzero

z (n:Nat) : Term n
z = λ $ λ $ var fzero

s (n:Nat) : Term n
s {n} = λ $ λ $ λ $ ss @ (nn @ ss @ zz)
  where
    nn : Term (3 + n)
    nn = var $ fsuc $ fsuc fzero
    ss : Term (3 + n)
    ss = var $ fsuc fzero
    zz : Term (3 + n)
    zz = var fzero

one (n:Nat) : Term n
one = s @ z

oneC (n:Nat) : Term n
oneC {n} = λ $ λ $ ss @ zz
  where
    ss : Term (2 + n)
    ss = var (fsuc fzero)
    zz : Term (2 + n)
    zz = var fzero

oneR (n:Nat) : one -->* oneC
oneR {n} = step1 `tccons'` step2
		 `tccons'` step3
		 `tccons'` tcrefl {oneC {n}}
  where
    one1 : Term n
    one1 = λ (λ (var (fsuc fzero) @ (z @ var (fsuc fzero) @ var fzero)))

    one2 : Term n
    one2 = λ (λ (var (fsuc fzero) @ (λ (var fzero) @ var fzero)))

    step1 : one --> one1
    step1 = β

    step2 : one1 --> one2
    step2 = rλ (rλ (rapp2 (rapp1 β)))

    step3 : one2 --> oneC
    step3 = rλ (rλ (rapp2 β))

two (n:Nat) : Term n
two = s @ one

twoC (n:Nat) : Term n
twoC {n} = λ $ λ $ ss @ (ss @ zz)
  where
    ss : Term (2 + n)
    ss = var (fsuc fzero)

    zz : Term (2 + n)
    zz = var fzero

-- twoR (n:Nat) : two -->* twoC
-- twoR = 

test : { v : Term 0; s @ oneC -->* v /\ _ -->* v }
test = confluence two (s @ oneC) _ (tcapp2 oneR) (tcunit β)

