
-- # import id
-- # import nat

{--------------------------------------------------------------------------
    Finite sets
 --------------------------------------------------------------------------}

data Fin : Nat -> Set where
  fzero	(n : Nat) : Fin (suc n)
  fsuc  (n : Nat) : Fin n -> Fin (suc n)

data NonZero : Nat -> Set where
  nonZero (n : Nat) : NonZero (suc n)

finNonZero (n:Nat) : Fin n -> NonZero n
finNonZero fzero    = nonZero
finNonZero (fsuc _) = nonZero

{--------------------------------------------------------------------------
    Thinning
 --------------------------------------------------------------------------}

infixr 25 `thin`, `thick`

thin (n : Nat) : Fin (suc n) -> Fin n -> Fin (suc n)
thin fzero    i		= fsuc i
thin (fsuc i) fzero	= fzero
thin (fsuc i) (fsuc j)	= fsuc (thin i j)

inj (n : Nat) : Fin n -> Fin (suc n)
inj fzero     = fzero
inj (fsuc i)  = fsuc (inj i)

data ThinView : (n : Nat) -> Fin n -> Fin n -> Set where
  hit (n : Nat)(i : Fin n)	  : ThinView n i i
  miss (n : Nat)(i : Fin (suc n)) : (j : Fin n) -> ThinView (suc n) i (thin i j)

thinView (n : Nat) : (i, j : Fin n) -> ThinView n i j
thinView fzero		 fzero	  = hit
thinView fzero		 (fsuc j) = miss j
thinView (fsuc i)	 fzero	with finNonZero i =>
	 (fsuc i)	 fzero	  nonZero = miss fzero
thinView (fsuc i)	 (fsuc j) with thinView i j =>
	 (fsuc i) (fsuc i)          hit	     = hit
	 (fsuc i) (fsuc (thin i j)) (miss j) = miss (fsuc j)

thinViewSuc0 (n : Nat) : (i : Fin (suc n)) ->
		      thinView (fsuc i) fzero == miss fzero
thinViewSuc0 fzero    = refl
thinViewSuc0 (fsuc _) = refl

thinViewHit (n : Nat) : (i : Fin n) -> thinView i i == hit
thinViewHit fzero     = refl
thinViewHit (fsuc i)  = rewrite thinViewHit i in refl

thinViewMiss (n : Nat) : (i : Fin (suc n)) -> (j : Fin n) ->
			 thinView i (thin i j) == miss j
thinViewMiss fzero j		= refl
thinViewMiss (fsuc i) fzero	= thinViewSuc0 i
thinViewMiss (fsuc i) (fsuc j)	= rewrite thinViewMiss i j in refl

{--------------------------------------------------------------------------
    Thicking
 --------------------------------------------------------------------------}

-- | thin (thin j i) (thick i j) == j
thick (n : Nat) : Fin n -> Fin (suc n) -> Fin n
thick i fzero	with finNonZero i => i fzero nonZero = fzero
thick fzero (fsuc j)  = j
thick (fsuc i) (fsuc j)	= fsuc (thick i j)

thickZero (n : Nat) : (i : Fin (suc n)) -> thick i fzero == fzero
thickZero fzero     = refl
thickZero (fsuc _)  = refl

thickThinThin (n : Nat) :
    (i : Fin n) -> (j : Fin (suc n)) ->
    thin (thin j i) (thick i j) == j
thickThinThin fzero fzero	= refl
thickThinThin (fsuc i) fzero	= refl
thickThinThin fzero (fsuc j)	= refl
thickThinThin (fsuc i) (fsuc j) = rewrite thickThinThin i j in refl

thickThin (n : Nat) :
    (i, j : Fin n) ->
    thick i (thin (inj i) j) == j
thickThin fzero j	    = refl
thickThin (fsuc i) fzero    = refl
thickThin (fsuc i) (fsuc j) = rewrite thickThin i j in refl

ThinViewThick (n : Nat) : Fin (suc n) -> Fin n -> Set
ThinViewThick i j = rewrite thickThinThin j i in thinView (thin i j) i == miss (thick j i)

thinViewThick (n : Nat) : (i : Fin (suc n)) -> (j : Fin n) ->
			  ThinViewThick i j
thinViewThick fzero fzero	= refl
thinViewThick fzero (fsuc j)	= refl
thinViewThick (fsuc i) fzero	= refl
thinViewThick (fsuc i) (fsuc j) = rewrite thinViewThick i j in refl

thinCommute (n : Nat) :
    (k : Fin n) -> (i : Fin (suc n)) -> (j : Fin (suc (suc n))) ->
    thin j (thin i k) == thin (thin j i) (thin (thick i j) k)
thinCommute  k	     fzero     fzero	= refl
thinCommute  k	    (fsuc i)   fzero	= refl
thinCommute  k	     fzero    (fsuc j)	= refl
thinCommute  fzero   (fsuc i)  (fsuc j)	= refl
thinCommute (fsuc k) (fsuc i)  (fsuc j)	= cong fsuc (thinCommute k i j)

