
-- # import Set
-- # import Rel
-- # import Prop

-- Setoids ----------------------------------------------------------------

sig Setoid : Type extends SET where
  eq : EquivalenceRel (mkSET el)

mkSetoid : (A:SET) -> EquivalenceRel A -> Setoid
mkSetoid A eq = struct el = El A; eq = eq

infix 9 ==

(==) (A:Setoid) : El A -> El A -> PROP
(==) {A} = A.eq.rel

-- Antisymmetric relations ------------------------------------------------

AntiSymmetric (A:Setoid) : Rel A -> PROP
AntiSymmetric R =
  ForAll $ \x -> ForAll $ \y ->
  R.rel x y  ==>  R.rel y x  ==>  x == y

sig AntiSymmetricRel (A:Setoid) : Type extends Rel where
  antisym : Prf (AntiSymmetric (mkRel rel))

antisymmetry (A:Setoid)(R:AntiSymmetricRel A)(x,y:El A) :
  Prf (R.rel x y ==> R.rel y x ==> x == y)
antisymmetry {_} {R} xRy yRx = R.antisym _ _ xRy yRx

-- Setoid relations (respecting the equality) -----------------------------

RespectsEq (A:Setoid) : Rel A -> PROP
RespectsEq R =
  ForAll $ \x -> ForAll $ \y -> ForAll $ \z -> ForAll $ \w ->
    x == y  ==>  z == w  ==>  R.rel x z  ==>  R.rel y w

sig SetoidRel (A:Setoid) : Type extends Rel where
  respeq : Prf (RespectsEq (mkRel rel))

respectsEq (A:Setoid)(R:SetoidRel A)(x,y,z,w:El A) :
  Prf (x == y ==> z == w ==> R.rel x z ==> R.rel y w)
respectsEq {_} {R} = R.respeq _ _ _ _

respectsEq1 (A:Setoid)(R:SetoidRel A)(x,y,z:El A) :
  Prf (x == y ==> R.rel x z ==> R.rel y z)
respectsEq1 {A} "x=y" = respectsEq "x=y" (reflexivity {A})

respectsEq2 (A:Setoid)(R:SetoidRel A)(x,y,z:El A) :
  Prf (y == z ==> R.rel x y ==> R.rel x z)
respectsEq2 {A} "y=z" = respectsEq (reflexivity {A}) "y=z"

-- Reflexive setoid relations ---------------------------------------------

sig ReflexiveSetoidRel (A:Setoid) : Type
  extends ReflexiveRel, SetoidRel where

setoidRefl (A:Setoid)(R:ReflexiveSetoidRel A)(x,y:El A) :
  Prf (x == y ==> R.rel x y)
setoidRefl {A} {R} "x=y" =
  respectsEq (reflexivity {A}) "x=y" (reflexivity {A})

-- Irreflexive relations --------------------------------------------------

sig IrreflexiveSetoidRel (A:Setoid) : Type
  extends IrreflexiveRel, SetoidRel where

setoidIrrefl (A:Setoid)(R:IrreflexiveSetoidRel A)(x,y:El A) :
  Prf (x == y ==> Not (R.rel x y))
setoidIrrefl {A} {R} {x} {y} "x=y" xRy =
    notE (irreflexivity {A}) yRy
  where
    yRy = respectsEq1 "x=y" xRy

